# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses
from typing import Any

from ansys.api.acp.v0 import (
    analysis_ply_pb2_grpc,
    cut_off_geometry_pb2_grpc,
    enum_types_pb2,
    imported_analysis_ply_pb2_grpc,
    imported_solid_model_pb2,
    imported_solid_model_pb2_grpc,
    layup_mapping_object_pb2_grpc,
    solid_element_set_pb2_grpc,
    solid_model_pb2,
)

from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from .._utils.typing_helper import PATH as _PATH
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.enum_wrapper import wrap_to_string_enum
from ._grpc_helpers.exceptions import wrap_grpc_errors
from ._grpc_helpers.mapping import (
    define_create_method,
    define_mutable_mapping,
    get_read_only_collection_property,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._mesh_data import solid_mesh_property
from ._solid_model_export import SolidModelExportMixin
from .analysis_ply import AnalysisPly
from .base import (
    CreatableTreeObject,
    IdTreeObject,
    TreeObjectAttributeWithCache,
    nested_grpc_object_property,
)
from .cut_off_geometry import CutOffGeometry
from .enums import (
    UnitSystemType,
    status_type_from_pb,
    unit_system_type_from_pb,
    unit_system_type_to_pb,
)
from .imported_analysis_ply import ImportedAnalysisPly
from .layup_mapping_object import LayupMappingObject
from .material import Material
from .object_registry import register
from .solid_element_set import SolidElementSet

__all__ = [
    "ImportedSolidModel",
    "SolidModelImportFormat",
    "ImportedSolidModelExportSettings",
    "ImportedSolidModelElementalData",
    "ImportedSolidModelNodalData",
]


SolidModelImportFormat, solid_model_import_format_to_pb, solid_model_import_format_from_pb = (
    wrap_to_string_enum(
        "SolidModelImportFormat",
        enum_types_pb2.FileFormat,
        module=__name__,
        value_converter=lambda val: val.lower().replace("_", ":"),
        doc="Options for the file format when importing a solid model.",
        explicit_value_list=(
            enum_types_pb2.FileFormat.ANSYS_H5,
            enum_types_pb2.FileFormat.ANSYS_CDB,
            enum_types_pb2.FileFormat.ANSYS_DAT,
        ),
    )
)


@dataclasses.dataclass
class ImportedSolidModelElementalData(ElementalData):
    """Represents elemental data for an imported solid model."""


@dataclasses.dataclass
class ImportedSolidModelNodalData(NodalData):
    """Represents nodal data for an imported solid model."""


@mark_grpc_properties
class ImportedSolidModelExportSettings(TreeObjectAttributeWithCache):
    """Defines the settings for exporting an imported solid model.

    Parameters
    ----------
    use_default_section_index :
        Use the default start index for sections.
    section_index :
        Custom start index for sections.
        Only used if ``use_default_section_index`` is False.
    use_default_coordinate_system_index :
        Use the default start index for coordinate systems.
    coordinate_system_index :
        Custom start index for coordinate systems.
        Only used if ``use_default_coordinate_system_index`` is False.
    use_default_material_index :
        Use the default start index for materials.
    material_index :
        Custom start index for materials.
        Only used if ``use_default_material_index`` is False.
    use_default_node_index :
        Use the default start index for nodes.
    node_index :
        Custom start index for nodes.
        Only used if ``use_default_node_index`` is False.
    use_default_element_index :
        Use the default start index for elements.
    element_index :
        Custom start index for elements.
        Only used if ``use_default_element_index`` is False.
    use_solsh_elements :
        When True, export linear layered elements as Solsh (Solid190).
    drop_hanging_nodes :
        When True, the hanging nodes of quadratic solid meshes are dropped.
    use_solid_model_prefix :
        Use the imported solid model name as a prefix for the exported file.

    """

    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        use_default_section_index: bool = True,
        section_index: int = 0,
        use_default_coordinate_system_index: bool = True,
        coordinate_system_index: int = 0,
        use_default_material_index: bool = True,
        material_index: int = 0,
        use_default_node_index: bool = True,
        node_index: int = 0,
        use_default_element_index: bool = True,
        element_index: int = 0,
        use_solsh_elements: bool = False,
        drop_hanging_nodes: bool = True,
        use_solid_model_prefix: bool = True,
        _parent_object: ImportedSolidModel | None = None,
        _pb_object: Any | None = None,
        _attribute_path: str | None = None,
    ):
        super().__init__(
            _parent_object=_parent_object,
            _pb_object=_pb_object,
            _attribute_path=_attribute_path,
        )
        # See comment on DropOffSettings.__init__ for the logic here.
        if _parent_object is None and _pb_object is None:
            self.use_default_section_index = use_default_section_index
            self.section_index = section_index
            self.use_default_coordinate_system_index = use_default_coordinate_system_index
            self.coordinate_system_index = coordinate_system_index
            self.use_default_material_index = use_default_material_index
            self.material_index = material_index
            self.use_default_node_index = use_default_node_index
            self.node_index = node_index
            self.use_default_element_index = use_default_element_index
            self.element_index = element_index
            self.use_solsh_elements = use_solsh_elements
            self.drop_hanging_nodes = drop_hanging_nodes
            self.use_solid_model_prefix = use_solid_model_prefix

    @classmethod
    def _create_default_pb_object(self) -> solid_model_pb2.ExportSettings:
        # See comment on DropOffSettings._create_default_pb_object
        return solid_model_pb2.ExportSettings()

    use_default_section_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_section_index"
    )
    section_index: ReadWriteProperty[int, int] = grpc_data_property("section_index")
    use_default_coordinate_system_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_coordinate_system_index"
    )
    coordinate_system_index: ReadWriteProperty[int, int] = grpc_data_property(
        "coordinate_system_index"
    )
    use_default_material_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_material_index"
    )
    material_index: ReadWriteProperty[int, int] = grpc_data_property("material_index")
    use_default_node_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_node_index"
    )
    node_index: ReadWriteProperty[int, int] = grpc_data_property("node_index")
    use_default_element_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_element_index"
    )
    element_index: ReadWriteProperty[int, int] = grpc_data_property("element_index")
    use_solsh_elements: ReadWriteProperty[bool, bool] = grpc_data_property("use_solsh_elements")
    drop_hanging_nodes: ReadWriteProperty[bool, bool] = grpc_data_property("drop_hanging_nodes")
    use_solid_model_prefix: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_solid_model_prefix"
    )


@mark_grpc_properties
@register
class ImportedSolidModel(SolidModelExportMixin, CreatableTreeObject, IdTreeObject):
    """Instantiate an imported solid model.

    Parameters
    ----------
    name :
        Name of the imported solid model.
    active :
        Inactive imported solid models are ignored in the analysis.
    format :
        Specifies the format of the file to be imported.
    unit_system :
        Specifies the unit system of the external solid mesh.
    external_path :
        Path of the file to be imported.
    delete_bad_elements :
        If True, run an element check and delete erroneous elements. Bad elements
        can falsify the mapping.
    warping_limit :
        Maximum allowable warping. Elements with a warping above this limit are
        deleted.
        Only used if ``delete_bad_elements`` is True.
    minimum_volume :
        Solid elements with a volume smaller or equal to this value are deleted.
        Only used if ``delete_bad_elements`` is True.
    cut_off_material :
        This material is assigned to the degenerated solid cut-off elements if
        ``cut_off_material_handling`` is set to ``GLOBAL`` in the fabric
        definition.
    export_settings :
        Defines the settings for exporting the imported solid model.

    """

    __slots__: Iterable[str] = tuple()
    _COLLECTION_LABEL = "imported_solid_models"
    _OBJECT_INFO_TYPE = imported_solid_model_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = imported_solid_model_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "ImportedSolidModel",
        active: bool = True,
        format: SolidModelImportFormat = "ansys:cdb",  # type: ignore
        unit_system: UnitSystemType = "from_file",
        external_path: _PATH = "",
        delete_bad_elements: bool = True,
        warping_limit: float = 0.4,
        minimum_volume: float = 0.0,
        cut_off_material: Material | None = None,
        export_settings: ImportedSolidModelExportSettings = ImportedSolidModelExportSettings(),
    ):
        super().__init__(name=name)
        self.active = active
        self.format = format
        self.unit_system = unit_system
        self.external_path = external_path
        self.delete_bad_elements = delete_bad_elements
        self.warping_limit = warping_limit
        self.minimum_volume = minimum_volume
        self.cut_off_material = cut_off_material
        self.export_settings = export_settings

    def _create_stub(self) -> imported_solid_model_pb2_grpc.ObjectServiceStub:
        return imported_solid_model_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    format = grpc_data_property(
        "properties.format",
        from_protobuf=solid_model_import_format_from_pb,
        to_protobuf=solid_model_import_format_to_pb,
    )
    unit_system = grpc_data_property(
        "properties.unit_system",
        from_protobuf=unit_system_type_from_pb,
        to_protobuf=unit_system_type_to_pb,
    )
    external_path: ReadWriteProperty[str, _PATH] = grpc_data_property(
        "properties.external_path", to_protobuf=str
    )
    delete_bad_elements: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.delete_bad_elements"
    )
    warping_limit: ReadWriteProperty[float, float] = grpc_data_property("properties.warping_limit")
    minimum_volume: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.minimum_volume"
    )
    cut_off_material = grpc_link_property("properties.cut_off_material", allowed_types=(Material,))
    export_settings = nested_grpc_object_property(
        "properties.export_settings", ImportedSolidModelExportSettings
    )

    solid_element_sets = get_read_only_collection_property(
        SolidElementSet, solid_element_set_pb2_grpc.ObjectServiceStub
    )

    elemental_data = elemental_data_property(ImportedSolidModelElementalData)
    nodal_data = nodal_data_property(ImportedSolidModelNodalData)

    create_cut_off_geometry = define_create_method(
        CutOffGeometry,
        func_name="create_cut_off_geometry",
        parent_class_name="ImportedSolidModel",
        module_name=__module__,
    )
    cut_off_geometries = define_mutable_mapping(
        CutOffGeometry, cut_off_geometry_pb2_grpc.ObjectServiceStub
    )

    create_layup_mapping_object = define_create_method(
        LayupMappingObject,
        func_name="create_layup_mapping_object",
        parent_class_name="ImportedSolidModel",
        module_name=__name__,
    )
    layup_mapping_objects = define_mutable_mapping(
        LayupMappingObject,
        layup_mapping_object_pb2_grpc.ObjectServiceStub,
    )

    analysis_plies = get_read_only_collection_property(
        AnalysisPly, analysis_ply_pb2_grpc.ObjectServiceStub, supported_since="25.2"
    )
    imported_analysis_plies = get_read_only_collection_property(
        ImportedAnalysisPly,
        imported_analysis_ply_pb2_grpc.ObjectServiceStub,
        supported_since="25.2",
    )

    def refresh(self, path: _PATH, format: SolidModelImportFormat | None = None) -> None:  # type: ignore
        """
        Re-import the solid model from the external file.

        Parameters
        ----------
        path :
            Path of the new input file.
        format :
            Switch format of the input file. Optional, uses the current format of the
            imported solid model if not specified.
        """
        if format is not None:
            self.format = format
        self.external_path = self._server_wrapper.auto_upload(path)
        with wrap_grpc_errors():
            self._get_stub().Refresh(  # type: ignore
                imported_solid_model_pb2.RefreshRequest(resource_path=self._resource_path)
            )

    def import_initial_mesh(self) -> None:
        """Import the solid mesh and its element sets."""
        with wrap_grpc_errors():
            self._get_stub().ImportInitialMesh(  # type: ignore
                imported_solid_model_pb2.ImportInitialMeshRequest(resource_path=self._resource_path)
            )

    solid_mesh = solid_mesh_property



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

from ansys.api.acp.v0 import (
    butt_joint_sequence_pb2_grpc,
    interface_layer_pb2_grpc,
    modeling_group_pb2,
    modeling_group_pb2_grpc,
    modeling_ply_pb2_grpc,
)

from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.mapping import define_create_method, define_mutable_mapping
from ._grpc_helpers.property_helper import mark_grpc_properties
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .butt_joint_sequence import ButtJointSequence
from .interface_layer import InterfaceLayer
from .modeling_ply import ModelingPly
from .object_registry import register

__all__ = ["ModelingGroup"]


@dataclasses.dataclass
class ModelingGroupElementalData(ElementalData):
    """Represents elemental data for a Modeling Group."""

    normal: VectorData | None = None


@dataclasses.dataclass
class ModelingGroupNodalData(NodalData):
    """Represents nodal data for a Modeling Group."""


@mark_grpc_properties
@register
class ModelingGroup(CreatableTreeObject, IdTreeObject):
    """Instantiate a modeling group.

    Parameters
    ----------
    name :
        Name of the modeling group.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "modeling_groups"
    _OBJECT_INFO_TYPE = modeling_group_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = modeling_group_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(self, *, name: str = "ModelingGroup"):
        super().__init__(name=name)

    def _create_stub(self) -> modeling_group_pb2_grpc.ObjectServiceStub:
        return modeling_group_pb2_grpc.ObjectServiceStub(self._channel)

    create_modeling_ply = define_create_method(
        ModelingPly,
        func_name="create_modeling_ply",
        parent_class_name="ModelingGroup",
        module_name=__module__,
    )
    modeling_plies = define_mutable_mapping(ModelingPly, modeling_ply_pb2_grpc.ObjectServiceStub)
    create_interface_layer = define_create_method(
        InterfaceLayer,
        func_name="create_interface_layer",
        parent_class_name="ModelingGroup",
        module_name=__module__,
    )
    interface_layers = define_mutable_mapping(
        InterfaceLayer, interface_layer_pb2_grpc.ObjectServiceStub
    )

    create_butt_joint_sequence = define_create_method(
        ButtJointSequence,
        func_name="create_butt_joint_sequence",
        parent_class_name="ModelingGroup",
        module_name=__module__,
    )
    butt_joint_sequences = define_mutable_mapping(
        ButtJointSequence, butt_joint_sequence_pb2_grpc.ObjectServiceStub
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property

    elemental_data = elemental_data_property(ModelingGroupElementalData)
    nodal_data = nodal_data_property(ModelingGroupNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable, Iterable
import typing
from typing import Any

from typing_extensions import Self

from ansys.api.acp.v0 import base_pb2, virtual_geometry_pb2, virtual_geometry_pb2_grpc

from ._grpc_helpers.edge_property_list import (
    GenericEdgePropertyType,
    define_add_method,
    define_edge_property_list,
)
from ._grpc_helpers.property_helper import (
    _exposed_grpc_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .cad_geometry import CADGeometry
from .enums import status_type_from_pb, virtual_geometry_dimension_from_pb
from .object_registry import register

if typing.TYPE_CHECKING:  # pragma: no cover
    from .cad_component import CADComponent


@mark_grpc_properties
class SubShape(GenericEdgePropertyType):
    """Represents a sub-shape of a virtual geometry."""

    _SUPPORTED_SINCE = "24.2"

    def __init__(self, cad_geometry: CADGeometry, path: str):
        self._callback_apply_changes: Callable[[], None] | None = None
        self.cad_geometry = cad_geometry
        self.path = path

    @_exposed_grpc_property
    def cad_geometry(self) -> CADGeometry:
        """Linked CAD geometry."""
        return self._cad_geometry

    @cad_geometry.setter
    def cad_geometry(self, value: CADGeometry) -> None:
        if not isinstance(value, CADGeometry):
            raise TypeError(f"Expected CADGeometry, got {type(value)}.")
        self._cad_geometry = value
        if self._callback_apply_changes:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def path(self) -> str:
        """Topological path of the sub-shape within the CAD geometry."""
        return self._path

    @path.setter
    def path(self, value: str) -> None:
        self._path = value
        if self._callback_apply_changes:
            self._callback_apply_changes()

    def _set_callback_apply_changes(self, callback_apply_changes: Callable[[], None]) -> None:
        self._callback_apply_changes = callback_apply_changes

    @classmethod
    def _from_pb_object(
        cls,
        parent_object: CreatableTreeObject,
        message: virtual_geometry_pb2.SubShape,
        apply_changes: Callable[[], None],
    ) -> SubShape:
        new_obj = cls(
            cad_geometry=CADGeometry._from_resource_path(
                message.cad_geometry, server_wrapper=parent_object._server_wrapper
            ),
            path=message.path,
        )
        new_obj._set_callback_apply_changes(apply_changes)
        return new_obj

    def _to_pb_object(self) -> virtual_geometry_pb2.SubShape:
        return virtual_geometry_pb2.SubShape(
            cad_geometry=self._cad_geometry._resource_path, path=self._path
        )

    def _check(self) -> bool:
        # Check for empty resource paths
        return bool(self._cad_geometry._resource_path.value)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return (
                self._cad_geometry._resource_path == other._cad_geometry._resource_path
                and self._path == other._path
            )

        return False

    def __repr__(self) -> str:
        return f"SubShape(cad_geometry={self._cad_geometry.__repr__()}, path={self._path})"

    def clone(self) -> Self:
        """Create a new unstored SubShape with the same properties."""
        return type(self)(cad_geometry=self.cad_geometry, path=self.path)


@mark_grpc_properties
@register
class VirtualGeometry(CreatableTreeObject, IdTreeObject):
    """Instantiate a Virtual Geometry.

    The virtual geometry can be created from a set of CAD components or from a set of SubShapes.
    Combining CAD Components and SubShapes is not allowed.

    Parameters
    ----------
    name :
        Name of the Virtual Geometry.
    cad_components :
        CAD Components that make up the virtual geometry.
    sub_shapes :
        SubShapes that make up the virtual geometry.

    """

    __slots__: Iterable[str] = tuple()
    _COLLECTION_LABEL = "virtual_geometries"
    _OBJECT_INFO_TYPE = virtual_geometry_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = virtual_geometry_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "VirtualGeometry",
        cad_components: Iterable[CADComponent] | None = None,
        sub_shapes: Iterable[SubShape] | None = None,
    ):
        super().__init__(
            name=name,
        )

        if cad_components is not None and sub_shapes is not None:
            raise ValueError("cad_components and sub_shapes cannot be set at the same time")

        if cad_components is None and sub_shapes is None:
            self.sub_shapes = []

        if cad_components is not None:
            self.set_cad_components(cad_components=cad_components)

        if sub_shapes is not None:
            self.sub_shapes = sub_shapes

    def _create_stub(self) -> virtual_geometry_pb2_grpc.ObjectServiceStub:
        return virtual_geometry_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    dimension = grpc_data_property_read_only(
        "properties.dimension", from_protobuf=virtual_geometry_dimension_from_pb
    )

    sub_shapes = define_edge_property_list(
        "properties.sub_shapes",
        SubShape,
    )
    add_sub_shape = define_add_method(
        SubShape,
        attribute_name="sub_shapes",
        func_name="add_sub_shape",
        parent_class_name="VirtualGeometry",
        module_name=__module__,
    )

    def set_cad_components(self, cad_components: Iterable[CADComponent]) -> None:
        """Set the sub-shapes to match the given CADComponents.

        Parameters
        ----------
        cad_components :
            The components which make up the virtual geometry.
        """

        # TODO: find a better / more generic way to do this
        def _get_parent(cad_component: CADComponent) -> CADGeometry:
            rp = "/".join(cad_component._resource_path.value.split("/")[:-2])
            return CADGeometry._from_resource_path(
                base_pb2.ResourcePath(value=rp), server_wrapper=cad_component._server_wrapper
            )

        sub_shapes = [
            SubShape(cad_geometry=_get_parent(component), path=component.path)
            for component in cad_components
        ]
        self.sub_shapes = sub_shapes



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from ansys.api.acp.v0 import material_pb2

from ..._grpc_helpers.property_helper import mark_grpc_properties
from ...base import TreeObject
from .base import _ConstantPropertySet, _VariablePropertySet
from .property_helper import (
    constant_material_grpc_data_property,
    variable_material_grpc_data_property,
)

__all__ = ["ConstantLaRCConstants", "VariableLaRCConstants"]


class _LaRCConstantsMixin:
    _DEFAULT_PB_PROPERTYSET_TYPE = material_pb2.LaRCConstantsPropertySet


@mark_grpc_properties
class ConstantLaRCConstants(_LaRCConstantsMixin, _ConstantPropertySet):
    """Constant LaRC failure criterion properties."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        fracture_angle_under_compression: float = 53.0,
        fracture_toughness_ratio: float = 0.0,
        fracture_toughness_mode_1: float = 0.0,
        fracture_toughness_mode_2: float = 0.0,
        thin_ply_thickness_limit: float = 0.0,
        _parent_object: TreeObject | None = None,
        _attribute_path: str | None = None,
    ):
        super().__init__(_parent_object=_parent_object, _attribute_path=_attribute_path)
        if _parent_object is not None:
            return
        self.fracture_angle_under_compression = fracture_angle_under_compression
        self.fracture_toughness_ratio = fracture_toughness_ratio
        self.fracture_toughness_mode_1 = fracture_toughness_mode_1
        self.fracture_toughness_mode_2 = fracture_toughness_mode_2
        self.thin_ply_thickness_limit = thin_ply_thickness_limit

    fracture_angle_under_compression = constant_material_grpc_data_property(
        "fracture_angle_under_compression"
    )
    fracture_toughness_ratio = constant_material_grpc_data_property("fracture_toughness_ratio")
    fracture_toughness_mode_1 = constant_material_grpc_data_property("fracture_toughness_mode_1")
    fracture_toughness_mode_2 = constant_material_grpc_data_property("fracture_toughness_mode_2")
    thin_ply_thickness_limit = constant_material_grpc_data_property("thin_ply_thickness_limit")


@mark_grpc_properties
class VariableLaRCConstants(_LaRCConstantsMixin, _VariablePropertySet):
    """Variable LaRC failure criterion properties."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    fracture_angle_under_compression = variable_material_grpc_data_property(
        "fracture_angle_under_compression"
    )
    fracture_toughness_ratio = variable_material_grpc_data_property("fracture_toughness_ratio")
    fracture_toughness_mode_1 = variable_material_grpc_data_property("fracture_toughness_mode_1")
    fracture_toughness_mode_2 = variable_material_grpc_data_property("fracture_toughness_mode_2")
    thin_ply_thickness_limit = variable_material_grpc_data_property("thin_ply_thickness_limit")



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Sequence
import pathlib
import typing

from ansys.api.acp.v0 import section_cut_pb2, section_cut_pb2_grpc

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from .._utils.path_to_str import path_to_str_checked
from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from .._utils.typing_helper import PATH as _PATH
from ._grpc_helpers.exceptions import wrap_grpc_errors
from ._grpc_helpers.linked_object_list import define_linked_object_list
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .element_set import ElementSet
from .enums import (
    ExtrusionType,
    IntersectionType,
    SectionCutCDBExportType,
    SectionCutType,
    extrusion_type_from_pb,
    extrusion_type_to_pb,
    intersection_type_from_pb,
    intersection_type_to_pb,
    section_cut_cdb_export_type_to_pb,
    section_cut_type_from_pb,
    section_cut_type_to_pb,
    status_type_from_pb,
)
from .object_registry import register

__all__ = ["SectionCut"]


@mark_grpc_properties
@register
class SectionCut(CreatableTreeObject, IdTreeObject):
    r"""Instantiate a Section Cut.

    Parameters
    ----------
    name :
        Name of the section cut.
    active :
        Inactive section cuts are not evaluated.
    origin :
        Defines the origin of the section cut plane.
    normal :
        Defines the normal vector of the section cut plane.
    in_plane_reference_direction1 :
        Defines the in-plane transverse direction of the beam. Used for the surface
        section cut and section cut measures.
    scope_entire_model :
        If True, the section cut is applied to the entire model. Otherwise, the
        section cut is applied only to the element sets specified in
        ``scope_element_sets``.
    scope_element_sets :
        The element sets to which the section cut is applied. Used only if
        ``scope_entire_model`` is False.
    extrusion_type :
        Determines the extrusion method used to create the section cut.
    scale_factor :
        Scale factor applied to the ply thicknesses.
    core_scale_factor :
        Scale factor applied to the core thicknesses.
    section_cut_type :
        Determines whether the section cut is extruded by modeling ply, production
        ply, or analysis ply.
    intersection_type :
        Determines the method used to compute a wire frame section cut. Used only
        if ``extrusion_type`` is ``ExtrusionType.WIRE_FRAME``.
    use_default_tolerance :
        If True, the segment tolerance is computed as 0.1\% of the averaged element size.
        Otherwise, the ``tolerance`` value is used.
        Used only if ``extrusion_type`` is ``ExtrusionType.SURFACE_NORMAL`` or
        ``ExtrusionType.SURFACE_SWEEP_BASED``.
    tolerance :
        Defines the minimum length of the segments. Segments shorter than this value
        are merged.
        Used only if ``extrusion_type`` is ``ExtrusionType.SURFACE_NORMAL`` or
        ``ExtrusionType.SURFACE_SWEEP_BASED``, and ``use_default_tolerance`` is
        False.
    use_default_interpolation_settings :
        If True, default interpolation settings are used by the sweep-based extrusion.
        Used only if ``extrusion_type`` is ``ExtrusionType.SURFACE_SWEEP_BASED``.
    search_radius :
        Search radius of the interpolation algorithm used in the sweep-based extrusion.
        Used only if ``extrusion_type`` is ``ExtrusionType.SURFACE_SWEEP_BASED`` and
        ``use_default_interpolation_settings`` is False.
    number_of_interpolation_points :
        Number of interpolation points of the interpolation algorithm used in the
        sweep-based extrusion.
        Used only if ``extrusion_type`` is ``ExtrusionType.SURFACE_SWEEP_BASED`` and
        ``use_default_interpolation_settings`` is False.

    """

    __slots__ = ()

    _COLLECTION_LABEL = "section_cuts"
    _OBJECT_INFO_TYPE = section_cut_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = section_cut_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "SectionCut",
        active: bool = True,
        origin: tuple[float, float, float] = (0.0, 0.0, 0.0),
        normal: tuple[float, float, float] = (0.0, 0.0, 1.0),
        in_plane_reference_direction1: tuple[float, float, float] = (1.0, 0.0, 0.0),
        scope_entire_model: bool = True,
        scope_element_sets: Sequence[ElementSet] = tuple(),
        extrusion_type: ExtrusionType = ExtrusionType.WIRE_FRAME,
        scale_factor: float = 1.0,
        core_scale_factor: float = 1.0,
        section_cut_type: SectionCutType = SectionCutType.MODELING_PLY_WISE,
        intersection_type: IntersectionType = IntersectionType.NORMAL_TO_SURFACE,
        use_default_tolerance: bool = True,
        tolerance: float = 0.0,
        use_default_interpolation_settings: bool = True,
        search_radius: float = 0.0,
        number_of_interpolation_points: int = 1,
    ) -> None:
        super().__init__(name=name)
        self.active = active
        self.origin = origin
        self.normal = normal
        self.in_plane_reference_direction1 = in_plane_reference_direction1
        self.scope_entire_model = scope_entire_model
        self.scope_element_sets = scope_element_sets
        self.extrusion_type = extrusion_type
        self.scale_factor = scale_factor
        self.core_scale_factor = core_scale_factor
        self.section_cut_type = section_cut_type
        self.intersection_type = intersection_type
        self.use_default_tolerance = use_default_tolerance
        self.tolerance = tolerance
        self.use_default_interpolation_settings = use_default_interpolation_settings
        self.search_radius = search_radius
        self.number_of_interpolation_points = number_of_interpolation_points

    def _create_stub(self) -> section_cut_pb2_grpc.ObjectServiceStub:
        return section_cut_pb2_grpc.ObjectServiceStub(self._channel)

    # general properties
    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")

    # position properties
    origin = grpc_data_property(
        "properties.origin", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )
    normal = grpc_data_property(
        "properties.normal", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )
    in_plane_reference_direction1 = grpc_data_property(
        "properties.in_plane_reference_direction1",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )

    # scoping properties
    scope_entire_model: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.scope_entire_model"
    )
    scope_element_sets = define_linked_object_list("properties.scope_element_sets", ElementSet)

    # extrusion properties
    extrusion_type = grpc_data_property(
        "properties.extrusion_type",
        from_protobuf=extrusion_type_from_pb,
        to_protobuf=extrusion_type_to_pb,
    )
    scale_factor: ReadWriteProperty[float, float] = grpc_data_property("properties.scale_factor")
    core_scale_factor: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.core_scale_factor"
    )
    section_cut_type = grpc_data_property(
        "properties.section_cut_type",
        from_protobuf=section_cut_type_from_pb,
        to_protobuf=section_cut_type_to_pb,
    )

    # wireframe properties
    intersection_type = grpc_data_property(
        "properties.intersection_type",
        from_protobuf=intersection_type_from_pb,
        to_protobuf=intersection_type_to_pb,
    )

    # surface properties - general
    use_default_tolerance: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_tolerance"
    )
    tolerance: ReadWriteProperty[float, float] = grpc_data_property("properties.tolerance")
    # surface properties - sweep-based
    use_default_interpolation_settings: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_interpolation_settings"
    )
    search_radius: ReadWriteProperty[float, float] = grpc_data_property("properties.search_radius")
    number_of_interpolation_points: ReadWriteProperty[int, int] = grpc_data_property(
        "properties.number_of_interpolation_points"
    )

    def export(
        self,
        path: _PATH,
        *,
        export_type: SectionCutCDBExportType = "mesh_only",
    ) -> None:
        """Export the section cut to a CDB file.

        Parameters
        ----------
        path :
            Path to the file where the section cut is saved.
        export_type :
            Determines what is exported to the CDB file. Options are:

            - ``"mesh_only"``: Only the mesh (elements and nodes) is exported.
            - ``"solid_model"``: The section cut is expanded into a slice of
              solid elements. In addition, the material properties are exported
              and the element coordinate systems are aligned with the fiber
              direction. This model can be used to compute the equivalent
              beam properties of the section cut.

        """
        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                self._get_stub().ExportToCDB(  # type: ignore
                    section_cut_pb2.ExportToCDBRequest(
                        resource_path=self._resource_path,
                        path=export_path,
                        export_type=typing.cast(
                            typing.Any, section_cut_cdb_export_type_to_pb(export_type)
                        ),
                    )
                )

    def export_becas_input(
        self,
        path: _PATH,
        *,
        export_strength_limits: bool = True,
    ) -> None:
        """Export the section cut to BECAS input files.

        Parameters
        ----------
        path :
            Path to the directory where the input files are saved.
        export_strength_limits :
            Determines whether strength limits are exported to the
            BECAS input files.

        """
        # The 'auto_download' context manager cannot be directly used here
        # because the export path is not a file but a directory. The BECAS
        # export produces multiple files, which we all need to download.
        # Currently (May '25), this is the only export that produces multiple
        # files at once. If there are more, we should either extend 'auto_download',
        # or create a new context manager for this purpose.
        export_path = pathlib.Path(
            self._server_wrapper.filetransfer_handler.to_export_path(path, is_directory=True)
        )
        expected_filenames = ["E2D.in", "EMAT.in", "MATPROPS.in", "N2D.in"]
        if export_strength_limits:
            expected_filenames.append("FAILMAT.in")

        with wrap_grpc_errors():
            self._get_stub().ExportToBECAS(  # type: ignore
                section_cut_pb2.ExportToBECASRequest(
                    resource_path=self._resource_path,
                    path=path_to_str_checked(export_path),
                    export_strength_limits=export_strength_limits,
                )
            )
        for filename in expected_filenames:
            self._server_wrapper.filetransfer_handler.download_file_if_autotransfer(
                export_path / filename, pathlib.Path(path) / filename
            )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helper functions for dealing with Resource Paths."""

__all__ = ["join"]


def join(*parts: str) -> str:
    """Join parts of a Resource Path into a single string.

    Join parts of a Resource Path by slashes, ensuring that there are no
    double slashes in the result. Leading and trailing slashes are removed.
    """
    return "/".join(parts).replace("//", "/").strip("/")


def to_parts(path: str) -> tuple[str, ...]:
    """Get a tuple representation of the given path."""
    return tuple(path.split("/"))


def common_path(*paths: str) -> str:
    """Get the partial path that all input paths have in common."""
    common_parts = []
    for path_parts in zip(*[to_parts(p) for p in paths]):
        first_path_part = path_parts[0]
        if all(part == first_path_part for part in path_parts):
            common_parts.append(first_path_part)
        else:
            break
    return "/".join(common_parts)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os

from ._tree_objects._grpc_helpers.mapping import Mapping
from ._tree_objects.base import TreeObjectBase
from ._tree_objects.model import Model
from ._utils.string_manipulation import replace_underscores_and_capitalize

__all__ = ["Node", "print_model", "get_model_tree"]


class Node:
    """A node in a tree representation of the model.

    Parameters
    ----------
    label:
        Label of the node.
    children:
        Children of the node.
    """

    def __init__(self, label: str, children: list["Node"] | None = None):
        self.label = label
        self.children: list["Node"] = children if children else []

    def __str__(self) -> str:
        return self.to_string(_prefix="", show_lines=False)

    def to_string(
        self,
        *,
        show_lines: bool = False,
        _prefix: str = "",
        _is_last_child: bool = True,
        _is_root: bool = True,
    ) -> str:
        """Convert the node to a string representation.

        Parameters
        ----------
        show_lines :
            Whether to show lines connecting the nodes.
        """
        if show_lines:
            elbow = "└── "
            line = "│   "
            tee = "├── "
            empty = "    "
        else:
            elbow = line = tee = empty = " " * 4
        if _is_root:
            res = _prefix + self.label + os.linesep
        else:
            res = _prefix + (elbow if _is_last_child else tee) + self.label + os.linesep
        for i, child in enumerate(self.children):
            if _is_root:
                new_prefix = _prefix
            elif _is_last_child:
                new_prefix = _prefix + empty
            else:
                new_prefix = _prefix + line
            res += child.to_string(
                _prefix=new_prefix,
                show_lines=show_lines,
                _is_last_child=(i == len(self.children) - 1),
                _is_root=False,
            )
        return res


def print_model(
    model: Model, *, hide_empty: bool = True, show_lines: bool = False, label_by_id: bool = False
) -> None:
    """Print a tree representation of the model.

    Parameters
    ----------
    model:
        pyACP model
    hide_empty :
        Whether to hide empty collections.
    show_lines :
        Whether to show lines connecting the nodes.
    label_by_id :
        Prefer the ID over the name for the label of a node, for objects that have both.
    """
    return print(
        get_model_tree(model, hide_empty=hide_empty, label_by_id=label_by_id).to_string(
            show_lines=show_lines
        )
    )


def get_model_tree(model: Model, *, hide_empty: bool = True, label_by_id: bool = False) -> Node:
    """Get a tree representation of the model.

    Returns the root node.

    Parameters
    ----------
    model :
        ACP model.
    hide_empty :
        Whether to hide empty collections.
    label_by_id :
        Prefer the ID over the name for the label of a node, for objects that have both.
    """
    return _get_model_tree_impl(obj=model, hide_empty=hide_empty, label_by_id=label_by_id)


def _get_model_tree_impl(obj: TreeObjectBase, *, hide_empty: bool, label_by_id: bool) -> Node:
    obj_node = Node(repr(_name_or_id(obj, label_by_id=label_by_id)))
    for attr_name in obj._GRPC_PROPERTIES:
        try:
            attr = getattr(obj, attr_name)
        except (AttributeError, RuntimeError):
            continue
        if isinstance(attr, Mapping):
            collection_node = Node(replace_underscores_and_capitalize(attr_name))
            obj_node.children.append(collection_node)
            for child_obj in attr.values():
                collection_node.children.append(
                    _get_model_tree_impl(child_obj, hide_empty=hide_empty, label_by_id=label_by_id)
                )
            if hide_empty and not collection_node.children:
                obj_node.children.pop()
    return obj_node


def _name_or_id(obj: TreeObjectBase, label_by_id: bool) -> str:
    if label_by_id:
        try:
            return obj.id  # type: ignore
        except AttributeError:
            return obj.name
    else:
        try:
            return obj.name
        except AttributeError:
            return obj.id  # type: ignore



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .density import ConstantDensity, VariableDensity
from .engineering_constants import ConstantEngineeringConstants, VariableEngineeringConstants
from .fabric_fiber_angle import ConstantFabricFiberAngle, VariableFabricFiberAngle
from .larc_constants import ConstantLaRCConstants, VariableLaRCConstants
from .puck_constants import ConstantPuckConstants, PuckMaterialType, VariablePuckConstants
from .strain_limits import ConstantStrainLimits, VariableStrainLimits
from .stress_limits import ConstantStressLimits, VariableStressLimits
from .tsai_wu_constants import ConstantTsaiWuConstants, VariableTsaiWuConstants
from .variable_property_set_attributes import FieldVariable, InterpolationOptions
from .woven_characterization import ConstantWovenCharacterization, VariableWovenCharacterization
from .woven_stress_limits import ConstantWovenStressLimits, VariableWovenStressLimits
from .wrapper import wrap_property_set

__all__ = [
    "ConstantDensity",
    "ConstantEngineeringConstants",
    "ConstantFabricFiberAngle",
    "ConstantLaRCConstants",
    "ConstantPuckConstants",
    "ConstantStrainLimits",
    "ConstantStressLimits",
    "ConstantTsaiWuConstants",
    "ConstantWovenCharacterization",
    "ConstantWovenStressLimits",
    "FieldVariable",
    "InterpolationOptions",
    "PuckMaterialType",
    "VariableDensity",
    "VariableEngineeringConstants",
    "VariableFabricFiberAngle",
    "VariableLaRCConstants",
    "VariablePuckConstants",
    "VariableStrainLimits",
    "VariableStressLimits",
    "VariableTsaiWuConstants",
    "VariableWovenCharacterization",
    "VariableWovenStressLimits",
    "wrap_property_set",
]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Collection
from typing import Any, overload

import numpy as np
import numpy.typing as npt

from ansys.api.acp.v0.array_types_pb2 import DoubleArray, Int32Array, IntArray
from ansys.api.acp.v0.mesh_query_pb2 import DataArray


def to_1D_double_array(data: Collection[float]) -> DoubleArray:
    """Convert a 1D collection of floats to a DoubleArray protobuf message."""
    return DoubleArray(shape=[len(data)], data=tuple(data))


def to_1D_int_array(data: Collection[int]) -> IntArray:
    """Convert a 1D collection of ints to a IntArray protobuf message."""
    return IntArray(shape=[len(data)], data=tuple(data))


def to_tuple_from_1D_array(array: IntArray | DoubleArray) -> tuple[Any, ...]:
    """Convert a 1D IntArray or DoubleArray protobuf message to a tuple."""
    if not len(array.shape) == 1:
        raise RuntimeError(f"Cannot convert {len(array.shape)}-dimensional array to tuple!")
    return tuple(array.data)


def to_ND_double_array_from_numpy_or_list(data: npt.NDArray[np.float64]) -> DoubleArray:
    """Convert a list or numpy array to a DoubleArray protobuf message."""
    data_np = np.array(data)
    return DoubleArray(shape=list(data_np.shape), data=data_np.flatten())


@overload
def to_numpy(array_pb: IntArray) -> npt.NDArray[np.int64]: ...


@overload
def to_numpy(array_pb: Int32Array) -> npt.NDArray[np.int32]: ...


@overload
def to_numpy(array_pb: DoubleArray) -> npt.NDArray[np.float64]: ...


def to_numpy(
    array_pb: IntArray | Int32Array | DoubleArray,
) -> npt.NDArray[np.int64] | npt.NDArray[np.int32] | npt.NDArray[np.float64]:
    """Convert a protubuf array message to a numpy array."""
    dtype = {
        IntArray: np.int64,
        Int32Array: np.int32,
        DoubleArray: np.float64,
    }[type(array_pb)]
    return np.array(array_pb.data, dtype=dtype).reshape(array_pb.shape)


def dataarray_to_numpy(
    array_pb: DataArray,
    dtype: type[np.int32] | type[np.int64] | type[np.float64],
) -> npt.NDArray[np.int64] | npt.NDArray[np.int32] | npt.NDArray[np.float64]:
    """Convert a DataArray protobuf message to a numpy array."""
    data_array_attribute = array_pb.WhichOneof("data")
    if data_array_attribute is None:
        raise RuntimeError("None of the 'DataArray' data attributes are set!")
    expected_data_array_attribute = {
        np.int64: "int_array",
        np.int32: "int32_array",
        np.float64: "double_array",
    }[dtype]
    if data_array_attribute != expected_data_array_attribute:
        raise RuntimeError(
            f"Wrong data attribute is set: expected '{expected_data_array_attribute}', "
            f"got '{data_array_attribute}'!"
        )
    return to_numpy(getattr(array_pb, data_array_attribute))



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

import numpy as np

from ansys.api.acp.v0 import analysis_ply_pb2, analysis_ply_pb2_grpc

from .._utils.property_protocols import ReadOnlyProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    ScalarData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property_read_only,
    grpc_link_property_read_only,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property, solid_mesh_property
from .base import IdTreeObject, ReadOnlyTreeObject
from .enums import status_type_from_pb
from .object_registry import register

__all__ = ["AnalysisPly", "AnalysisPlyElementalData", "AnalysisPlyNodalData"]


@dataclasses.dataclass
class AnalysisPlyElementalData(ElementalData):
    """Represents elemental data for a Analysis Ply."""

    normal: VectorData | None = None
    orientation: VectorData | None = None
    reference_direction: VectorData | None = None
    fiber_direction: VectorData | None = None
    draped_fiber_direction: VectorData | None = None
    transverse_direction: VectorData | None = None
    draped_transverse_direction: VectorData | None = None
    thickness: ScalarData[np.float64] | None = None
    relative_thickness_correction: ScalarData[np.float64] | None = None
    design_angle: ScalarData[np.float64] | None = None
    shear_angle: ScalarData[np.float64] | None = None
    draped_fiber_angle: ScalarData[np.float64] | None = None
    draped_transverse_angle: ScalarData[np.float64] | None = None
    area: ScalarData[np.float64] | None = None
    price: ScalarData[np.float64] | None = None
    volume: ScalarData[np.float64] | None = None
    mass: ScalarData[np.float64] | None = None
    offset: ScalarData[np.float64] | None = None
    material_1_direction: VectorData | None = None
    cog: VectorData | None = None


@dataclasses.dataclass
class AnalysisPlyNodalData(NodalData):
    """Represents nodal data for an Analysis Ply."""

    ply_offset: VectorData | None = None


@mark_grpc_properties
@register
class AnalysisPly(ReadOnlyTreeObject, IdTreeObject):
    """Instantiate an Analysis Ply.

    Parameters
    ----------
    name: str
        The name of the Analysis Ply.
    material: Material
        Material of the Analysis ply.
    angle: float
        Angle of the Analysis ply in degrees.
    thickness: float
        Thickness of the Analysis ply
    active_in_post_mode: bool
        If False, deactivates the failure analysis for this ply during post-processing.

    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "analysis_plies"
    _OBJECT_INFO_TYPE = analysis_ply_pb2.ObjectInfo
    _SUPPORTED_SINCE = "24.2"

    def _create_stub(self) -> analysis_ply_pb2_grpc.ObjectServiceStub:
        return analysis_ply_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    material = grpc_link_property_read_only("properties.material")
    angle: ReadOnlyProperty[float] = grpc_data_property_read_only("properties.angle")
    thickness: ReadOnlyProperty[float] = grpc_data_property_read_only("properties.thickness")
    active_in_post_mode: ReadOnlyProperty[bool] = grpc_data_property_read_only(
        "properties.active_in_post_mode"
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property
    solid_mesh = solid_mesh_property
    elemental_data = elemental_data_property(AnalysisPlyElementalData)
    nodal_data = nodal_data_property(AnalysisPlyNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from ansys.api.acp.v0 import material_pb2

from ..._grpc_helpers.property_helper import mark_grpc_properties
from ...base import TreeObject
from .base import _ConstantPropertySet, _VariablePropertySet
from .property_helper import (
    constant_material_grpc_data_property,
    variable_material_grpc_data_property,
)

__all__ = ["ConstantWovenStressLimits", "VariableWovenStressLimits"]


class _WovenStressLimitsMixin:
    _DEFAULT_PB_PROPERTYSET_TYPE = material_pb2.OrthotropicStressLimitsPropertySet


@mark_grpc_properties
class ConstantWovenStressLimits(_WovenStressLimitsMixin, _ConstantPropertySet):
    """Constant stress limits property set for woven materials."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        Xc: float = 0.0,
        Yc: float = 0.0,
        Zc: float = 0.0,
        Xt: float = 0.0,
        Yt: float = 0.0,
        Zt: float = 0.0,
        Sxy: float = 0.0,
        Syz: float = 0.0,
        Sxz: float = 0.0,
        _parent_object: TreeObject | None = None,
        _attribute_path: str | None = None,
    ):
        super().__init__(_parent_object=_parent_object, _attribute_path=_attribute_path)
        if _parent_object is not None:
            return
        self.Xc = Xc
        self.Yc = Yc
        self.Zc = Zc
        self.Xt = Xt
        self.Yt = Yt
        self.Zt = Zt
        self.Sxy = Sxy
        self.Syz = Syz
        self.Sxz = Sxz

    Xc = constant_material_grpc_data_property("Xc")
    Yc = constant_material_grpc_data_property("Yc")
    Zc = constant_material_grpc_data_property("Zc")
    Xt = constant_material_grpc_data_property("Xt")
    Yt = constant_material_grpc_data_property("Yt")
    Zt = constant_material_grpc_data_property("Zt")
    Sxy = constant_material_grpc_data_property("Sxy")
    Syz = constant_material_grpc_data_property("Syz")
    Sxz = constant_material_grpc_data_property("Sxz")


@mark_grpc_properties
class VariableWovenStressLimits(_WovenStressLimitsMixin, _VariablePropertySet):
    """Variable stress limits property set for woven materials."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    Xc = variable_material_grpc_data_property("Xc")
    Yc = variable_material_grpc_data_property("Yc")
    Zc = variable_material_grpc_data_property("Zc")
    Xt = variable_material_grpc_data_property("Xt")
    Yt = variable_material_grpc_data_property("Yt")
    Zt = variable_material_grpc_data_property("Zt")
    Sxy = variable_material_grpc_data_property("Sxy")
    Syz = variable_material_grpc_data_property("Syz")
    Sxz = variable_material_grpc_data_property("Sxz")



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Callable, Mapping
import types
from typing import Any

__all__ = ["wrap_to_string_enum"]

from ansys.acp.core._utils.typing_helper import StrEnum

# mypy doesn't understand this dynamically created Enum, so we have to
# fall back to 'Any'.
_StrEnumT = Any


def wrap_to_string_enum(
    class_name: str,
    proto_enum: Any,
    module: str,
    *,
    key_converter: Callable[[str], str] = lambda val: val,
    value_converter: Callable[[str], str] = lambda val: val.lower(),
    doc: str,
    explicit_value_list: tuple[int, ...] | None = None,
    extra_aliases: Mapping[str, tuple[str, str]] = types.MappingProxyType({}),
    normalization_function: Callable[[str], str] = lambda val: val.lower(),
) -> tuple[_StrEnumT, Callable[[_StrEnumT], int], Callable[[int], _StrEnumT]]:
    """Create a string Enum with the same keys as the given protobuf Enum.

    Values of the enum are the keys, converted to lowercase.

    Parameters
    ----------
    key_converter :
        A callable which converts the protobuf field names to the string enum field names.
    value_converter :
        A callable which converts the protobuf field names to the string enum values.
    doc :
        The docstring of the enum.
    explicit_value_list :
        A list of values that should be included in the enum. If None, all values are included.
    extra_aliases :
        Allows defining additional fields in the enum which correspond to the same protobuf value.
        The keys are the primary enum field values, and the values are tuples of the alias field name
        and the alias field value.
        Note that the alias will not be used when converting from the protobuf value to the string
        enum: the primary field name will be used instead.
    normalization_function :
        A callable which is applied to strings before converting them to protobuf. This means that
        any string which will result in an enum value will be accepted, for example if it has the
        wrong capitalization.

    Returns
    -------
    :
        A tuple containing
        - the string enum
        - a conversion function from string enum to protobuf
        - a conversion function from protobuf to string enum
    """
    fields = []
    to_pb_conversion_dict: dict[Any, int] = {}
    from_pb_conversion_dict: dict[int, Any] = {}
    for key, pb_value in proto_enum.items():
        if explicit_value_list is not None:
            if pb_value not in explicit_value_list:
                continue
        enum_key = key_converter(key)
        enum_value = value_converter(key)
        fields.append((enum_key, enum_value))
        to_pb_conversion_dict[enum_value] = pb_value
        from_pb_conversion_dict[pb_value] = enum_value
    for primary_enum_value, (alias_enum_key, alias_enum_value) in extra_aliases.items():
        fields.append((alias_enum_key, alias_enum_value))
        to_pb_conversion_dict[alias_enum_value] = to_pb_conversion_dict[primary_enum_value]

    res_enum: _StrEnumT = StrEnum(class_name, fields, module=module)  # type: ignore
    res_enum.__doc__ = doc

    def to_pb_conversion_func(val: _StrEnumT) -> int:
        val = normalization_function(val)
        val = res_enum(val)  # generates a nicer error if 'val' is not a valid enum value
        return to_pb_conversion_dict[val]

    def from_pb_conversion_func(val: int) -> _StrEnumT:
        return res_enum(from_pb_conversion_dict[val])

    return (
        res_enum,
        to_pb_conversion_func,
        from_pb_conversion_func,
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

from ansys.api.acp.v0 import (
    lookup_table_1d_column_pb2_grpc,
    lookup_table_1d_pb2,
    lookup_table_1d_pb2_grpc,
)

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from ._grpc_helpers.mapping import define_create_method, define_mutable_mapping
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import status_type_from_pb
from .lookup_table_1d_column import LookUpTable1DColumn
from .object_registry import register

__all__ = ["LookUpTable1D"]


@mark_grpc_properties
@register
class LookUpTable1D(CreatableTreeObject, IdTreeObject):
    """Instantiate a 1D Look-Up Table.

    The one-dimensional look-up table is defined along an axis. The locations
    of the data points along the axis are defined by the ``Location`` column,
    which:

    * contains scalar data. When the column contains ``N`` data points (rows),
      it has shape ``(N, )``.
    * is automatically created when the look-up table is instantiated
    * cannot be deleted or renamed

    When the length of the ``Location`` column is changed, the data of the
    other columns is automatically either truncated, or extended with ``NaN``
    values to fit.

    Parameters
    ----------
    origin :
        Origin of the axis w.r.t. which the look-up table locations are
        defined.
    direction :
        Direction of the axis w.r.t. which the look-up table locations are
        defined.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "lookup_tables_1d"
    _OBJECT_INFO_TYPE = lookup_table_1d_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = lookup_table_1d_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "LookUpTable1D",
        origin: tuple[float, float, float] = (0.0, 0.0, 0.0),
        direction: tuple[float, float, float] = (0.0, 0.0, 0.0),
    ):
        super().__init__(name=name)

        self.origin = origin
        self.direction = direction

    def _create_stub(self) -> lookup_table_1d_pb2_grpc.ObjectServiceStub:
        return lookup_table_1d_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    origin = grpc_data_property(
        "properties.origin", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )
    direction = grpc_data_property(
        "properties.direction", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )

    create_column = define_create_method(
        LookUpTable1DColumn,
        func_name="create_column",
        parent_class_name="LookUpTable1D",
        module_name=__module__,
    )
    columns = define_mutable_mapping(
        LookUpTable1DColumn, lookup_table_1d_column_pb2_grpc.ObjectServiceStub
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

from ansys.api.acp.v0 import interface_layer_pb2, interface_layer_pb2_grpc

from .._utils.property_protocols import ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.linked_object_list import (
    define_linked_object_list,
    define_polymorphic_linked_object_list,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .element_set import ElementSet
from .enums import status_type_from_pb
from .object_registry import register
from .oriented_selection_set import OrientedSelectionSet


@dataclasses.dataclass
class InterfaceLayerElementalData(ElementalData):
    """Represents elemental data for a Modeling Ply."""

    normal: VectorData | None = None


@dataclasses.dataclass
class InterfaceLayerNodalData(NodalData):
    """Represents nodal data for a Modeling Ply."""

    ply_offset: VectorData | None = None


@mark_grpc_properties
@register
class InterfaceLayer(CreatableTreeObject, IdTreeObject):
    """Instantiate an interface layer.

    The interface layer is a separation layer in the stacking sequence. It can be
    used to analyze the crack growth of existing cracks. They can also be used to
    define contacts zones between two layers.
    The topology is defined with an interface layer in ACP, while all other fracture
    settings need to be specified in the downstream analysis (MAPDL or Mechanical).

    Parameters
    ----------
    name :
        Name of the interface layer.
    global_ply_nr :
        Global ply number for the stacking sequence.
    active :
        Inactive interface layers are ignored in ACP and the downstream analysis.
    oriented_selection_sets :
        Oriented Selection Set for the expansion of the interface layer.
    open_area_sets :
        Defines the initial crack of a Virtual Crack Closure Technique (VCCT) layer.
        Can contain ``OrientedSelectionSet`` and ``ElementSet`` objects.
    """

    __slots__: Iterable[str] = tuple()
    _COLLECTION_LABEL = "interface_layers"
    _OBJECT_INFO_TYPE = interface_layer_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = interface_layer_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "InterfaceLayer",
        global_ply_nr: int = 0,
        active: bool = True,
        oriented_selection_sets: Iterable[OrientedSelectionSet] = (),
        open_area_sets: Iterable[ElementSet | OrientedSelectionSet] = (),
    ):
        super().__init__(name=name)
        self.global_ply_nr = global_ply_nr
        self.active = active
        self.oriented_selection_sets = oriented_selection_sets
        self.open_area_sets = open_area_sets

    def _create_stub(self) -> interface_layer_pb2_grpc.ObjectServiceStub:
        return interface_layer_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    global_ply_nr: ReadWriteProperty[int, int] = grpc_data_property("properties.global_ply_nr")
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    oriented_selection_sets = define_linked_object_list(
        "properties.oriented_selection_sets", OrientedSelectionSet
    )
    open_area_sets = define_polymorphic_linked_object_list(
        "properties.open_area_sets", allowed_types=(ElementSet, OrientedSelectionSet)
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property

    elemental_data = elemental_data_property(InterfaceLayerElementalData)
    nodal_data = nodal_data_property(InterfaceLayerNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from typing import Any

from typing_extensions import Self

from ansys.api.acp.v0 import stackup_pb2, stackup_pb2_grpc

from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from ._grpc_helpers.edge_property_list import (
    GenericEdgePropertyType,
    define_add_method,
    define_edge_property_list,
)
from ._grpc_helpers.property_helper import (
    _exposed_grpc_property,
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import (
    CutOffMaterialHandling,
    DrapingMaterialModel,
    DropOffMaterialHandling,
    SymmetryType,
    cut_off_material_type_from_pb,
    cut_off_material_type_to_pb,
    draping_material_type_from_pb,
    draping_material_type_to_pb,
    drop_off_material_type_from_pb,
    drop_off_material_type_to_pb,
    status_type_from_pb,
    symmetry_type_from_pb,
    symmetry_type_to_pb,
)
from .fabric import Fabric
from .material import Material
from .object_registry import register

__all__ = ["Stackup", "FabricWithAngle"]


@mark_grpc_properties
class FabricWithAngle(GenericEdgePropertyType):
    """Defines a fabric of a stackup.

    Parameters
    ----------
    fabric :
        Link to an existing fabric.
    angle :
        Orientation angle in degree of the fabric with respect to the reference direction.

    """

    _SUPPORTED_SINCE = "24.2"

    def __init__(self, fabric: Fabric, angle: float = 0.0):
        self._callback_apply_changes: Callable[[], None] | None = None
        self.fabric = fabric
        self.angle = angle

    @_exposed_grpc_property
    def fabric(self) -> Fabric:
        """Linked fabric."""
        return self._fabric

    @fabric.setter
    def fabric(self, value: Fabric) -> None:
        if not isinstance(value, Fabric):
            raise TypeError(f"Expected a Fabric, got {type(value)}")
        self._fabric = value
        if self._callback_apply_changes:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def angle(self) -> float:
        """Orientation angle in degree of the fabric with respect to the reference direction."""
        return self._angle

    @angle.setter
    def angle(self, value: float) -> None:
        self._angle = value
        if self._callback_apply_changes:
            self._callback_apply_changes()

    def _set_callback_apply_changes(self, callback_apply_changes: Callable[[], None]) -> None:
        self._callback_apply_changes = callback_apply_changes

    @classmethod
    def _from_pb_object(
        cls,
        parent_object: CreatableTreeObject,
        message: stackup_pb2.FabricWithAngle,
        apply_changes: Callable[[], None],
    ) -> FabricWithAngle:
        new_obj = cls(
            fabric=Fabric._from_resource_path(message.fabric, parent_object._channel),
            angle=message.angle,
        )
        new_obj._set_callback_apply_changes(apply_changes)
        return new_obj

    def _to_pb_object(self) -> stackup_pb2.FabricWithAngle:
        return stackup_pb2.FabricWithAngle(fabric=self.fabric._resource_path, angle=self.angle)

    def _check(self) -> bool:
        # Check for empty resource paths
        return bool(self.fabric._resource_path.value)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.fabric._resource_path == other.fabric._resource_path
                and self.angle == other.angle
            )

        return False

    def __repr__(self) -> str:
        return f"FabricWithAngle(fabric={self.fabric.__repr__()}, angle={self.angle})"

    def clone(self) -> Self:
        """Create a new unstored FabricWithAngle with the same properties."""
        return type(self)(fabric=self.fabric, angle=self.angle)


@mark_grpc_properties
@register
class Stackup(CreatableTreeObject, IdTreeObject):
    """Instantiate a Stackup.

    Parameters
    ----------
    name :
        Name of the stackup.
    symmetry :
        Whether the stackup is odd or even symmetrical, or none.
    topdown :
        The first fabric in the list is placed first in the mold if topdown is true.
    fabrics :
        List of fabrics with angles which build the stackup.
    area_price :
        Price per area of the stackup.
    drop_off_material_handling :
        Defines the material of drop-off elements in the solid model extrusion.
    drop_off_material :
        Specify the material of drop-off elements in the solid model.
    cut_off_material_handling :
        Defines the material of cut-off elements in solid models if cut-off geometries are active.
    cut_off_material :
        Define the cut-off material if a ply with this material is shaped by a cut-off geometry.
    draping_material_model :
        Specifies the draping model of the fabric.
    draping_ud_coefficient :
        Set the draping coefficient of the uni-directional draping model. Must be in the range of 0 to 1.

    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "stackups"
    _OBJECT_INFO_TYPE = stackup_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = stackup_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "Stackup",
        symmetry: SymmetryType = "no_symmetry",
        topdown: bool = True,
        fabrics: Sequence[FabricWithAngle] = tuple(),
        area_price: float = 0.0,
        drop_off_material_handling: DropOffMaterialHandling = "global",
        drop_off_material: Material | None = None,
        cut_off_material: Material | None = None,
        cut_off_material_handling: CutOffMaterialHandling = "computed",
        draping_material_model: DrapingMaterialModel = "woven",
        draping_ud_coefficient: float = 0.0,
    ):
        super().__init__(name=name)

        self.symmetry = SymmetryType(symmetry)
        self.topdown = topdown
        self.area_price = area_price
        self.fabrics = fabrics
        self.drop_off_material_handling = DropOffMaterialHandling(drop_off_material_handling)
        self.drop_off_material = drop_off_material
        self.cut_off_material_handling = CutOffMaterialHandling(cut_off_material_handling)
        self.cut_off_material = cut_off_material
        self.draping_material_model = DrapingMaterialModel(draping_material_model)
        self.draping_ud_coefficient = draping_ud_coefficient

    def _create_stub(self) -> stackup_pb2_grpc.ObjectServiceStub:
        return stackup_pb2_grpc.ObjectServiceStub(self._channel)

    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    thickness: ReadOnlyProperty[float] = grpc_data_property_read_only("properties.thickness")
    area_weight: ReadOnlyProperty[float] = grpc_data_property_read_only("properties.area_weight")

    symmetry = grpc_data_property(
        "properties.symmetry",
        from_protobuf=symmetry_type_from_pb,
        to_protobuf=symmetry_type_to_pb,
    )
    topdown: ReadWriteProperty[bool, bool] = grpc_data_property("properties.topdown")
    area_price: ReadWriteProperty[float, float] = grpc_data_property("properties.area_price")
    ignore_for_postprocessing: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.ignore_for_postprocessing"
    )

    drop_off_material_handling = grpc_data_property(
        "properties.drop_off_material_handling",
        from_protobuf=drop_off_material_type_from_pb,
        to_protobuf=drop_off_material_type_to_pb,
    )
    drop_off_material = grpc_link_property("properties.drop_off_material", allowed_types=Material)
    cut_off_material_handling = grpc_data_property(
        "properties.cut_off_material_handling",
        from_protobuf=cut_off_material_type_from_pb,
        to_protobuf=cut_off_material_type_to_pb,
    )
    cut_off_material = grpc_link_property("properties.cut_off_material", allowed_types=Material)
    draping_material_model = grpc_data_property(
        "properties.draping_material_model",
        from_protobuf=draping_material_type_from_pb,
        to_protobuf=draping_material_type_to_pb,
    )
    draping_ud_coefficient: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.draping_ud_coefficient"
    )

    fabrics = define_edge_property_list("properties.fabrics", FabricWithAngle)
    add_fabric = define_add_method(
        FabricWithAngle,
        attribute_name="fabrics",
        func_name="add_fabric",
        parent_class_name="Stackup",
        module_name=__module__,
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

from ansys.api.acp.v0 import material_pb2, material_pb2_grpc

from ..._utils.property_protocols import ReadOnlyProperty
from .._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from ..base import CreatableTreeObject, IdTreeObject
from ..enums import PlyType, ply_type_from_pb, ply_type_to_pb, status_type_from_pb
from ..object_registry import register
from .property_sets import (
    ConstantDensity,
    ConstantEngineeringConstants,
    ConstantFabricFiberAngle,
    ConstantLaRCConstants,
    ConstantPuckConstants,
    ConstantStrainLimits,
    ConstantStressLimits,
    ConstantTsaiWuConstants,
    ConstantWovenCharacterization,
    ConstantWovenStressLimits,
    VariableDensity,
    VariableEngineeringConstants,
    VariableFabricFiberAngle,
    VariableLaRCConstants,
    VariablePuckConstants,
    VariableStrainLimits,
    VariableStressLimits,
    VariableTsaiWuConstants,
    VariableWovenCharacterization,
    VariableWovenStressLimits,
    wrap_property_set,
)


@mark_grpc_properties
@register
class Material(CreatableTreeObject, IdTreeObject):
    """Instantiate a Material.

    .. note::

        The active material property sets are determined by the ``ply_type``. For
        example, ``puck_constants`` are only active when the ``ply_type`` is either
        :attr:`.PlyType.REGULAR` or :attr:`.PlyType.WOVEN`.

        The inactive property sets can still be *set*, but may not be accessible:

        - For **stored** materials, inactive property sets always return ``None``.
        - For **unstored** materials, inactive property sets can be accessed. They
          return ``None`` after storing.

    Parameters
    ----------
    name :
        Name of the Material.
    ply_type :
        Define the type of material such as core, uni-directional (regular), woven, or isotropic.
    density :
        Define the density of the material.
    engineering_constants :
        Define the material stiffness (Young's moduli, Poisson ratios and shear moduli).
    stress_limits :
        Define the stress limits for the evaluation of failure criteria with ansys.dpf.composites.
    strain_limits :
        Define the strain limits for the evaluation of failure criteria with ansys.dpf.composites.
    puck_constants :
        Define the puck constants for the evaluation of the Puck criterion with ansys.dpf.composites.
    woven_characterization :
        Define the puck constants of woven fabrics for the evaluation of the Puck
        criterion with ansys.dpf.composites.
    woven_puck_constants_1 :
        Define the puck constants of woven fabrics for the evaluation of the Puck
        criterion with ansys.dpf.composites.
    woven_puck_constants_2 :
        Define the puck constants of woven fabrics for the evaluation of the Puck
        criterion with ansys.dpf.composites.
    woven_puck_constants_2 :
        Define the puck constants of woven fabrics for the evaluation of the Puck
        criterion with ansys.dpf.composites.
    woven_stress_limits_1 :
        Define the puck constants of woven fabrics for the evaluation of the Puck
        criterion with ansys.dpf.composites.
    woven_stress_limits_2 :
        Define the puck constants of woven fabrics for the evaluation of the Puck
        criterion with ansys.dpf.composites.
    tsai_wu_constants :
        Define the Tsai-Wu constants for the evaluation of the Tsai-Wu criterion
        with ansys.dpf.composites.
    larc_constants :
        Define the LaRC constants for the evaluation of the LaRC criterion with
        ansys.dpf.composites.
    fabric_fiber_angle :
        Define the rotation angle between the material coordinate system and the fiber direction.
        Only used for shear dependent material properties which are provided by Material Designer.
    """

    _pb_object: material_pb2.ObjectInfo
    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "materials"
    _OBJECT_INFO_TYPE = material_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = material_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "Material",
        ply_type: PlyType = "undefined",
        density: ConstantDensity | None = None,
        engineering_constants: ConstantEngineeringConstants | None = None,
        stress_limits: ConstantStressLimits | None = None,
        strain_limits: ConstantStrainLimits | None = None,
        puck_constants: ConstantPuckConstants | None = None,
        woven_characterization: ConstantWovenCharacterization | None = None,
        woven_puck_constants_1: ConstantPuckConstants | None = None,
        woven_puck_constants_2: ConstantPuckConstants | None = None,
        woven_stress_limits_1: ConstantWovenStressLimits | None = None,
        woven_stress_limits_2: ConstantWovenStressLimits | None = None,
        tsai_wu_constants: ConstantTsaiWuConstants | None = None,
        larc_constants: ConstantLaRCConstants | None = None,
        fabric_fiber_angle: ConstantFabricFiberAngle | None = None,
    ):
        super().__init__(name=name)

        self.ply_type = ply_type
        self.density = density or ConstantDensity()
        self.engineering_constants = engineering_constants or ConstantEngineeringConstants()
        self.stress_limits = stress_limits
        self.strain_limits = strain_limits
        self.puck_constants = puck_constants
        self.woven_characterization = woven_characterization
        self.woven_puck_constants_1 = woven_puck_constants_1
        self.woven_puck_constants_2 = woven_puck_constants_2
        self.woven_stress_limits_1 = woven_stress_limits_1
        self.woven_stress_limits_2 = woven_stress_limits_2
        self.tsai_wu_constants = tsai_wu_constants
        self.larc_constants = larc_constants
        self.fabric_fiber_angle = fabric_fiber_angle

    density = wrap_property_set("density", ConstantDensity, VariableDensity)
    engineering_constants = wrap_property_set(
        "engineering_constants", ConstantEngineeringConstants, VariableEngineeringConstants
    )
    stress_limits = wrap_property_set("stress_limits", ConstantStressLimits, VariableStressLimits)
    strain_limits = wrap_property_set("strain_limits", ConstantStrainLimits, VariableStrainLimits)
    puck_constants = wrap_property_set(
        "puck_constants", ConstantPuckConstants, VariablePuckConstants
    )
    woven_characterization = wrap_property_set(
        "woven_characterization", ConstantWovenCharacterization, VariableWovenCharacterization
    )
    woven_puck_constants_1 = wrap_property_set(
        "woven_puck_constants_1", ConstantPuckConstants, VariablePuckConstants
    )
    woven_puck_constants_2 = wrap_property_set(
        "woven_puck_constants_2", ConstantPuckConstants, VariablePuckConstants
    )
    woven_stress_limits_1 = wrap_property_set(
        "woven_stress_limits_1", ConstantWovenStressLimits, VariableWovenStressLimits
    )
    woven_stress_limits_2 = wrap_property_set(
        "woven_stress_limits_2", ConstantWovenStressLimits, VariableWovenStressLimits
    )
    tsai_wu_constants = wrap_property_set(
        "tsai_wu_constants", ConstantTsaiWuConstants, VariableTsaiWuConstants
    )
    larc_constants = wrap_property_set(
        "larc_constants", ConstantLaRCConstants, VariableLaRCConstants
    )
    fabric_fiber_angle = wrap_property_set(
        "fabric_fiber_angle", ConstantFabricFiberAngle, VariableFabricFiberAngle
    )

    def _create_stub(self) -> material_pb2_grpc.ObjectServiceStub:
        return material_pb2_grpc.ObjectServiceStub(self._channel)

    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    ext_id: ReadOnlyProperty[str] = grpc_data_property_read_only("properties.ext_id")

    ply_type = grpc_data_property(
        "properties.ply_type",
        from_protobuf=ply_type_from_pb,
        to_protobuf=ply_type_to_pb,
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, MutableSequence
import inspect
import sys
import textwrap
from typing import Any, Concatenate, Protocol, TypeVar, cast, overload

from google.protobuf.message import Message
from typing_extensions import ParamSpec, Self

from .._object_cache import ObjectCacheMixin, constructor_with_cache
from ..base import CreatableTreeObject
from .property_helper import _exposed_grpc_property, _wrap_doc, grpc_data_getter, grpc_data_setter
from .protocols import GrpcObjectBase

__all__ = [
    "EdgePropertyList",
    "define_edge_property_list",
    "define_add_method",
    "GenericEdgePropertyType",
]


class GenericEdgePropertyType(GrpcObjectBase, Protocol):
    """Protocol for the definition of ACP edge properties such as FabricWithAngle."""

    def __init__(self, *kwargs: Any) -> None: ...

    @classmethod
    def _from_pb_object(
        cls,
        parent_object: CreatableTreeObject,
        message: Any,
        callback_apply_changes: Callable[[], None],
    ) -> Self: ...

    def _to_pb_object(self) -> Message: ...

    def _check(self) -> bool: ...

    def _set_callback_apply_changes(self, callback_apply_changes: Callable[[], None]) -> None: ...

    def clone(self) -> Self:
        """Create a new unstored object with the same properties."""
        raise NotImplementedError


ValueT = TypeVar("ValueT", bound=GenericEdgePropertyType)


class EdgePropertyList(ObjectCacheMixin, MutableSequence[ValueT]):
    """Wrap graph edges of a specific type.

    Wraps links between objects of a specific type, for instance FabricWithAngle
    of a stackup. This object handles the conversion of the protobuf messages to the
    corresponding Python object and vice-versa.
    self._object_list holds the Python object. The python objects are stored
    to support editing in-place editing. This is achieved by passing a callback
    function to the python object.

    The edges (and parent object) is updated
    as soon as an item or one of its properties are changed.

    Note: sort is not implemented because the model definition often depends
    on the order.

    The LinkedObjectList should be used for graph edges without a specific type.
    For instance, element sets of an oriented element set.
    """

    __slots__ = (
        "_object_list_store",
        "_parent_object",
        "_parent_was_stored",
        "_object_type",
        "_attribute_name",
        "_name",
        "_object_constructor",
    )

    @classmethod
    @constructor_with_cache(
        # NOTE greschd Feb'23:
        # We use the parent object's id() as part of the cache key since
        # the LinkedObjectList keeps its parent alive. This means the
        # id() does not change (and cannot be reused) as long as the
        # cache entry is alive.
        # This is somewhat incidental, but checked (indirectly) by the
        # object permanence tests. If we want to get rid of this, we
        # would instead need to find a way to handle the case where the
        # parent object is being stored.
        key_getter=lambda *args, parent_object, attribute_name, **kwargs: (
            id(parent_object),
            attribute_name,
        ),
        raise_on_invalid_key=True,
    )
    def _initialize_with_cache(
        cls: type[Self],
        *,
        parent_object: CreatableTreeObject,
        object_type: type[ValueT],
        attribute_name: str,
        from_pb_constructor: Callable[[CreatableTreeObject, Message, Callable[[], None]], ValueT],
    ) -> Self:
        return cls(
            _parent_object=parent_object,
            _object_type=object_type,
            _attribute_name=attribute_name,
            _from_pb_constructor=from_pb_constructor,
        )

    @staticmethod
    def _cache_key_valid(key: Any) -> bool:
        try:
            (parent_object_id, attribute_name) = key
            if not attribute_name:
                return False
            if not isinstance(parent_object_id, int):
                return False
            return True
        except Exception:
            return False

    def __init__(
        self,
        *,
        _parent_object: CreatableTreeObject,
        _object_type: type[ValueT],
        _attribute_name: str,
        _from_pb_constructor: Callable[[CreatableTreeObject, Message, Callable[[], None]], ValueT],
    ) -> None:
        self._parent_object = _parent_object
        self._parent_was_stored = self._parent_object._is_stored
        self._object_type = _object_type
        self._attribute_name = _attribute_name
        self._name = _attribute_name.split(".")[-1]

        self._object_constructor: Callable[[Message], ValueT] = (
            lambda pb_object: _from_pb_constructor(
                self._parent_object, pb_object, self._apply_changes
            )
        )
        if self._parent_object._is_stored:
            self._object_list_store = self._get_object_list_from_parent()
        else:
            # Cannot instantiate the objects if the server is not available
            self._object_list_store = []

    @property
    def _object_list(self) -> list[ValueT]:
        if self._parent_object._is_stored and not self._parent_was_stored:
            # There are two scenarios when the parent object becomes
            # stored:
            # - The _object_list already contained some values. In this case, we
            #   simply keep it, and make a (inexhaustive) check that the size
            #   matches.
            # - The parent object was default-constructed and then its _pb_object
            #   was then replaced (e.g. by a call to _from_object_info). In this
            #   case, the empty '_object_list' is no longer reflecting the actual
            #   state, and needs to be replaced.
            #   In general, we don't replace the _object_list to ensure any references
            #   to its elements are correctly updated (and retain the ability to
            #   update the object itself), but here it's not a concern since this
            #   only happens within the constructor.
            if self._object_list_store:
                assert len(self._object_list_store) == len(self._get_object_list_from_parent())
            else:
                self._object_list_store = self._get_object_list_from_parent()
            self._parent_was_stored = True
        return self._object_list_store

    def _get_object_list_from_parent(self) -> list[ValueT]:
        obj_list = []
        for item in grpc_data_getter(self._attribute_name, from_protobuf=list)(self._parent_object):
            obj_list.append(self._object_constructor(item))
        return obj_list

    def _set_object_list(self, items: list[ValueT]) -> None:
        """Set the object list on the parent AND updates the internal object list."""
        pb_obj_list = []
        for item in items:
            if not isinstance(item, self._object_type):
                raise TypeError(
                    f"Expected items of type {self._object_type}, got type {type(item)} instead. "
                    f"Item: {item}."
                )
            if not item._check():
                raise RuntimeError("Cannot initialize incomplete object.")
            pb_obj_list.append(item._to_pb_object())
            # update callback in case item was copied from another tree object
            # or if it is a new object
            item._set_callback_apply_changes(self._apply_changes)
        grpc_data_setter(self._attribute_name, to_protobuf=lambda x: x)(
            self._parent_object, pb_obj_list
        )
        # keep object list in sync with the backend. This is needed for the in-place editing
        self._object_list_store = items

    def __len__(self) -> int:
        return len(self._object_list)

    @overload
    def __getitem__(self, index: int) -> ValueT: ...

    @overload
    def __getitem__(self, index: slice) -> list[ValueT]: ...

    def __getitem__(self, index: int | slice) -> ValueT | list[ValueT]:
        obj_list = self._object_list[index]
        if not isinstance(obj_list, list):
            assert isinstance(obj_list, self._object_type)
            return obj_list
        return [item for item in obj_list]

    @overload
    def __setitem__(self, key: int, value: ValueT) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[ValueT]) -> None: ...

    def __setitem__(self, key: int | slice, value: ValueT | Iterable[ValueT]) -> None:
        # TODO: convert / check
        obj_list = self._object_list
        if isinstance(value, Iterable):
            if not isinstance(key, slice):
                raise TypeError("Cannot assign to a single index with an iterable.")
            obj_list[key] = [item for item in value]
        else:
            if not isinstance(key, int):
                raise TypeError("Cannot assign to a slice with a single object.")
            obj_list[key] = value
        self._set_object_list(obj_list)

    def __delitem__(self, key: int | slice) -> None:
        obj_list = self._object_list
        del obj_list[key]
        self._set_object_list(obj_list)

    def __iter__(self) -> Iterator[ValueT]:
        obj_list = self._object_list
        yield from (item for item in obj_list)

    def __reversed__(self) -> Iterator[ValueT]:
        obj_list = self._object_list
        yield from (item for item in reversed(obj_list))

    def __contains__(self, item: object) -> bool:
        return item in self._object_list

    def append(self, value: ValueT) -> None:
        """Append an object to the list.

        Parameters
        ----------
        value :
            The object to append.
        """
        obj_list = self._object_list
        obj_list.append(value)
        self._set_object_list(obj_list)

    def count(self, value: ValueT) -> int:
        """Count the number of occurrences of an object in the list.

        Parameters
        ----------
        value : ValueT
            The object to count.
        """
        return self._object_list.count(value)

    def index(self, value: ValueT, start: int = 0, stop: int = sys.maxsize) -> int:
        """Return the index of the first occurrence of an object in the list.

        Parameters
        ----------
        value :
            The object to search for.
        start :
            The index to start searching from.
        stop :
            The index to stop searching at.
        """
        return self._object_list.index(value, start, stop)

    def extend(self, iterable: Iterable[ValueT]) -> None:
        """Extend the list by appending all the items from the iterable.

        Parameters
        ----------
        iterable :
            The iterable to append.
        """
        obj_list = self._object_list
        obj_list.extend([it for it in iterable])
        self._set_object_list(obj_list)

    def insert(self, index: int, value: ValueT) -> None:
        """Insert an object at a given position.

        Parameters
        ----------
        index :
            The index to insert at.
        value :
            The object to insert.
        """
        obj_list = self._object_list
        obj_list.insert(index, value)
        self._set_object_list(obj_list)

    def pop(self, index: int = -1) -> ValueT:
        """Remove and return the object at the given index.

        Parameters
        ----------
        index :
            The index of the object to remove.
        """
        obj_list = self._object_list
        obj = obj_list.pop(index)
        self._set_object_list(obj_list)
        return obj

    def remove(self, value: ValueT) -> None:
        """Remove the first occurrence of an object from the list.

        Parameters
        ----------
        value :
            The object to remove.
        """
        obj_list = self._object_list
        obj_list.remove(value)
        self._set_object_list(obj_list)

    def reverse(self) -> None:
        """Reverse the list in-place."""
        self._set_object_list(list(reversed(self._object_list)))

    def _apply_changes(self) -> None:
        """Apply changes to the list.

        Use to support in-place modification.
        This function applies the changes if someone edits one entry of the list.
        """
        self._set_object_list(self._object_list)

    """
    It does not make to sense to sort them because the model depends on the order
    def sort(
        self, *, key: Callable[[ValueT], Any] = lambda obj: obj.__lt__, reverse: bool = False
    ) -> None:
        pb_object_list = self._get_pb_object_list()
        evaluated_key_list = [key(self._object_constructor(obj)) for obj in pb_object_list]
        idx_list = np.argsort(evaluated_key_list)
        if reverse:
            idx_list = idx_list[::-1]
        pb_object_list = list(np.array(pb_object_list)[idx_list])
        self._set_pb_object_list(pb_object_list)

    """

    def __eq__(self, other: Any) -> Any:
        return list(self) == other

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}[{self._object_type.__name__}] "
            + f"with parent {self._parent_object!r}, entries {self._object_list!r}>"
        )


def define_edge_property_list(
    attribute_name: str,
    value_type: type[GenericEdgePropertyType],
    *,
    doc: str | None = None,
) -> Any:
    """Define a list of linked tree objects with link properties."""

    def getter(
        self: CreatableTreeObject, *, check_stored: bool = True
    ) -> EdgePropertyList[GenericEdgePropertyType]:
        if check_stored and not self._is_stored:
            raise RuntimeError(
                f"Cannot access property {attribute_name.split('.')[-1]} on an object that is not stored."
            )
        return EdgePropertyList._initialize_with_cache(
            parent_object=self,
            object_type=value_type,
            attribute_name=attribute_name,
            from_pb_constructor=value_type._from_pb_object,
        )

    def setter(self: CreatableTreeObject, value: list[GenericEdgePropertyType]) -> None:
        # allow wholesale replacement on unstored objects
        getter(self, check_stored=False)[:] = value

    return _wrap_doc(_exposed_grpc_property(getter).setter(setter), doc=doc)


P = ParamSpec("P")
ParentT = TypeVar("ParentT", bound=CreatableTreeObject)


def define_add_method(
    value_type: Callable[P, ValueT],
    *,
    attribute_name: str,
    func_name: str,
    parent_class_name: str,
    module_name: str,
) -> Callable[Concatenate[ParentT, P], ValueT]:
    """Define a method to add a linked tree object with link properties."""

    def inner(self: ParentT, /, *args: P.args, **kwargs: P.kwargs) -> ValueT:
        edge_prop_list = cast(EdgePropertyList[ValueT], getattr(self, attribute_name))
        edge_prop_list.append(value_type(*args, **kwargs))
        return edge_prop_list[-1]

    inner.__doc__ = textwrap.dedent(
        f"""\
        Add a {value_type.__name__} to the {parent_class_name}.

        """
    )
    found_parameters = False
    if value_type.__doc__ is not None:
        doc_lines = textwrap.dedent(value_type.__doc__).splitlines()
        if "Parameters" in doc_lines:
            doc_lines = doc_lines[doc_lines.index("Parameters") :]
            inner.__doc__ += "\n".join(doc_lines)
            found_parameters = True
    if not found_parameters:
        inner.__doc__ += "See :class:`" + value_type.__name__ + "` for the available parameters.\n"

    parameters = [inspect.signature(inner).parameters["self"]]
    parameters.extend(inspect.signature(value_type).parameters.values())
    inner.__signature__ = inspect.Signature(parameters, return_annotation=value_type)  # type: ignore

    inner.__name__ = func_name
    inner.__qualname__ = f"{parent_class_name}.{func_name}"
    inner.__module__ = module_name
    return inner



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

from ansys.api.acp.v0 import boolean_selection_rule_pb2, boolean_selection_rule_pb2_grpc

from .._utils.property_protocols import ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.edge_property_list import define_add_method, define_edge_property_list
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .enums import status_type_from_pb
from .linked_selection_rule import LinkedSelectionRule
from .object_registry import register

# Workaround: these imports are needed to make sphinx_autodoc_typehints understand
# the inherited members of the Elemental- and NodalData classes.
import numpy as np  # noqa: F401 isort:skip
from ._elemental_or_nodal_data import ScalarData  # noqa: F401 isort:skip

__all__ = [
    "BooleanSelectionRule",
    "BooleanSelectionRuleElementalData",
    "BooleanSelectionRuleNodalData",
]


@dataclasses.dataclass
class BooleanSelectionRuleElementalData(ElementalData):
    """Represents elemental data for a Boolean Selection Rule."""

    normal: VectorData | None = None


@dataclasses.dataclass
class BooleanSelectionRuleNodalData(NodalData):
    """Represents nodal data for a Boolean Selection Rule."""


@mark_grpc_properties
@register
class BooleanSelectionRule(CreatableTreeObject, IdTreeObject):
    """Instantiate a Boolean Selection Rule.

    Parameters
    ----------
    name :
        Name of the Boolean Selection Rule.
    selection_rules :

    include_rule :
        Include or exclude area in rule. Setting this to ``False``
        inverts the selection.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "boolean_selection_rules"
    _OBJECT_INFO_TYPE = boolean_selection_rule_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = boolean_selection_rule_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "BooleanSelectionrule",
        selection_rules: Iterable[LinkedSelectionRule] = (),
        include_rule: bool = True,
    ):
        super().__init__(name=name)
        self.selection_rules = selection_rules
        self.include_rule = include_rule

    def _create_stub(self) -> boolean_selection_rule_pb2_grpc.ObjectServiceStub:
        return boolean_selection_rule_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    selection_rules = define_edge_property_list("properties.selection_rules", LinkedSelectionRule)
    add_selection_rule = define_add_method(
        LinkedSelectionRule,
        attribute_name="selection_rules",
        func_name="add_selection_rule",
        parent_class_name="BooleanSelectionRule",
        module_name=__module__,
    )

    include_rule: ReadWriteProperty[bool, bool] = grpc_data_property("properties.include_rule_type")

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property
    # selection rules don't have solid mesh data

    elemental_data = elemental_data_property(BooleanSelectionRuleElementalData)
    nodal_data = nodal_data_property(BooleanSelectionRuleNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Union, get_args

from ansys.api.acp.v0 import sensor_pb2, sensor_pb2_grpc

from .._utils.array_conversions import to_tuple_from_1D_array
from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from ._grpc_helpers.linked_object_list import define_polymorphic_linked_object_list
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .element_set import ElementSet
from .enums import SensorType, sensor_type_from_pb, sensor_type_to_pb, status_type_from_pb
from .fabric import Fabric
from .imported_solid_model import ImportedSolidModel
from .modeling_ply import ModelingPly
from .object_registry import register
from .oriented_selection_set import OrientedSelectionSet
from .solid_model import SolidModel
from .stackup import Stackup
from .sublaminate import SubLaminate

__all__ = ["Sensor"]


_LINKABLE_ENTITY_TYPES = Union[
    Fabric,
    Stackup,
    SubLaminate,
    ElementSet,
    OrientedSelectionSet,
    ModelingPly,
    SolidModel,
    ImportedSolidModel,
]


@mark_grpc_properties
@register
class Sensor(CreatableTreeObject, IdTreeObject):
    """Instantiate a Sensor.

    Parameters
    ----------
    name :
        Name of the sensor.
    sensor_type :
        Type of sensor: The sensor can be scoped by area, material, plies,
        or solid model.
    active :
        Inactive sensors are not evaluated.
    entities :
        List of entities which define the sensor's scope.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "sensors"
    _OBJECT_INFO_TYPE = sensor_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = sensor_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "Sensor",
        sensor_type: SensorType = SensorType.SENSOR_BY_AREA,
        active: bool = True,
        entities: Sequence[_LINKABLE_ENTITY_TYPES] = (),
    ):
        super().__init__(name=name)
        self.active = active
        self.entities = entities
        self.sensor_type = sensor_type

    def _create_stub(self) -> sensor_pb2_grpc.ObjectServiceStub:
        return sensor_pb2_grpc.ObjectServiceStub(self._channel)

    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    sensor_type = grpc_data_property(
        "properties.sensor_type", from_protobuf=sensor_type_from_pb, to_protobuf=sensor_type_to_pb
    )
    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    entities = define_polymorphic_linked_object_list(
        "properties.entities", allowed_types=get_args(_LINKABLE_ENTITY_TYPES)
    )

    covered_area: ReadOnlyProperty[float | None] = grpc_data_property_read_only(
        "properties.covered_area",
        check_optional=True,
        doc=(
            "The surface area of a selected Element Set / Oriented Selection Set, "
            "or the tooling surface area that is covered by the composite layup of "
            "the selected Material or Modeling Ply."
        ),
    )
    modeling_ply_area: ReadOnlyProperty[float | None] = grpc_data_property_read_only(
        "properties.modeling_ply_area",
        check_optional=True,
        doc="The surface area of all Modeling Plies of the selected entity.",
    )
    production_ply_area: ReadOnlyProperty[float | None] = grpc_data_property_read_only(
        "properties.production_ply_area",
        check_optional=True,
        doc="The surface area of all production plies of the selected entity.",
    )
    price: ReadOnlyProperty[float | None] = grpc_data_property_read_only(
        "properties.price",
        check_optional=True,
        doc=(
            "The price for the composite layup of the selected entity. The price "
            "per area is defined on the Fabrics or Stackups."
        ),
    )
    weight: ReadOnlyProperty[float | None] = grpc_data_property_read_only(
        "properties.weight", check_optional=True, doc="The mass of the selected entity."
    )
    center_of_gravity = grpc_data_property_read_only(
        "properties.center_of_gravity",
        from_protobuf=to_tuple_from_1D_array,
        check_optional=True,
        doc="The center of gravity of the selected entity in the global coordinate system.",
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import os

from packaging import version

from ansys.tools.local_product_launcher.config import get_launch_mode_for
from ansys.tools.local_product_launcher.interface import FALLBACK_LAUNCH_MODE_NAME
from ansys.tools.local_product_launcher.launch import launch_product

from .acp_instance import (
    ACPInstance,
    FileTransferHandler,
    FileTransferStrategy,
    LocalFileTransferStrategy,
    RemoteFileTransferStrategy,
)
from .common import ControllableServerProtocol, LaunchMode, ServerKey
from .connect import ConnectLaunchConfig
from .direct import DirectLaunchConfig
from .docker_compose import DockerComposeLaunchConfig

__all__ = ["launch_acp"]


def launch_acp(
    config: DirectLaunchConfig | DockerComposeLaunchConfig | ConnectLaunchConfig | None = None,
    launch_mode: LaunchMode | None = None,
    timeout: float | None = 30.0,
    auto_transfer_files: bool = True,
) -> ACPInstance[ControllableServerProtocol]:
    """Launch an ACP instance.

    Launch the ACP gRPC server with the given configuration. If no
    configuration is provided, the configured default is used.

    .. warning::

        Do not execute this function with untrusted input parameters.
        See the :ref:`security guide<security_launch_acp>` for details.

    Parameters
    ----------
    config :
        The configuration used for launching ACP. If unspecified, the
        default for the given launch mode is used.
    launch_mode :
        Specifies which ACP launcher is used. One of ``direct``,
        ``docker_compose``, or ``connect``. If unspecified, the
        configured default is used. If no default is configured,
        ``direct`` is used.
    timeout :
        Timeout to wait until ACP responds. If ``None`` is specified,
        the check that ACP has started is skipped.
    auto_transfer_files :
        Determines whether input and output files are automatically
        transferred (up- or downloaded) to the server. If ``True``,
        files are automatically transferred, and all paths in the
        import or export methods are *local* paths. If ``False``,
        file transfer needs to be handled manually, and the paths
        are relative to the server working directory.
        If the ``launch_mode`` is ``"direct"``, this only has an
        effect if the current working directory is changed after
        launching the server.

    Returns
    -------
    :
        ACP instance which can be used to control the server, and
        instantiate objects on the server.
    """
    launch_mode_evaluated = get_launch_mode_for(product_name="ACP", launch_mode=launch_mode)
    server_instance: ControllableServerProtocol = launch_product(
        product_name="ACP", config=config, launch_mode=launch_mode_evaluated
    )
    # The fallback launch mode for ACP is the direct launch mode.
    if launch_mode_evaluated in (LaunchMode.DIRECT, FALLBACK_LAUNCH_MODE_NAME):
        filetransfer_strategy: FileTransferStrategy = LocalFileTransferStrategy(os.getcwd())
        is_remote = False
    elif launch_mode_evaluated in (LaunchMode.DOCKER_COMPOSE, LaunchMode.CONNECT):
        filetransfer_strategy = RemoteFileTransferStrategy(
            channel_getter=lambda: server_instance.channels[ServerKey.FILE_TRANSFER],
        )
        is_remote = True
    else:
        raise ValueError("Invalid launch mode for ACP: " + str(launch_mode_evaluated))

    acp = ACPInstance(
        server=server_instance,
        filetransfer_handler=FileTransferHandler(
            filetransfer_strategy, auto_transfer_files=auto_transfer_files
        ),
        is_remote=is_remote,
    )
    if timeout is not None:
        acp._server.wait(timeout=timeout)
        # We can only check the server version after the server has started;
        # if the timeout is set to 'None', we skip this check.
        MIN_VERSION = "24.2"
        if version.parse(acp.server_version) < version.parse(MIN_VERSION):
            raise RuntimeError(
                f"ACP version {acp.server_version} is not supported. "
                f"Please use ACP version {MIN_VERSION} or later."
            )
    return acp



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Collection, Iterable

from ansys.api.acp.v0 import edge_set_pb2, edge_set_pb2_grpc

from .._utils.array_conversions import to_1D_double_array, to_1D_int_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .element_set import ElementSet
from .enums import EdgeSetType, edge_set_type_from_pb, edge_set_type_to_pb, status_type_from_pb
from .object_registry import register


@mark_grpc_properties
@register
class EdgeSet(CreatableTreeObject, IdTreeObject):
    """Instantiate an edge set.

    Parameters
    ----------
    name :
        Name of the edge set.
    edge_set_type :
        Determines how the edge set is defined. Can be one of:

        * :attr:`.EdgeSetType.BY_REFERENCE`: define the edge set using an :class:`.ElementSet`.
        * :attr:`.EdgeSetType.BY_NODES`: define the edge set using a list of node labels.

    defining_node_labels :
        Labels of the nodes in the edge set.
        This parameter only applies when ``edge_set_type`` is :attr:`.EdgeSetType.BY_NODES`.
    element_set :
        The boundary of this element set defines the initial
        edge set.
        This parameter only applies when ``edge_set_type`` is :attr:`.EdgeSetType.BY_REFERENCE`.
    limit_angle :
        The edge set is cropped if the angle between two element edges exceeds this limits (in degrees).
        Use ``-1.`` to disable cropping.
        This parameter only applies when ``edge_set_type`` is :attr:`.EdgeSetType.BY_REFERENCE`.
    origin :
        Defines the starting point of the edge set.
        This parameter only applies when ``edge_set_type`` is :attr:`.EdgeSetType.BY_REFERENCE`.
    """

    __slots__: Iterable[str] = tuple()
    _COLLECTION_LABEL = "edge_sets"
    _OBJECT_INFO_TYPE = edge_set_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = edge_set_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "EdgeSet",
        edge_set_type: EdgeSetType = EdgeSetType.BY_REFERENCE,
        defining_node_labels: Collection[int] = tuple(),
        element_set: ElementSet | None = None,
        limit_angle: float = -1.0,
        origin: tuple[float, float, float] = (0.0, 0.0, 0.0),
    ):
        super().__init__(
            name=name,
        )
        self.edge_set_type = edge_set_type
        self.defining_node_labels = defining_node_labels
        self.element_set = element_set
        self.limit_angle = limit_angle
        self.origin = origin

    def _create_stub(self) -> edge_set_pb2_grpc.ObjectServiceStub:
        return edge_set_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")

    edge_set_type = grpc_data_property(
        "properties.edge_set_type",
        from_protobuf=edge_set_type_from_pb,
        to_protobuf=edge_set_type_to_pb,
    )
    defining_node_labels = grpc_data_property(
        "properties.defining_node_labels",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_int_array,
    )
    element_set = grpc_link_property("properties.element_set", allowed_types=ElementSet)
    limit_angle: ReadWriteProperty[float, float] = grpc_data_property("properties.limit_angle")
    origin = grpc_data_property(
        "properties.origin", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing_extensions import Self

from ansys.api.acp.v0 import material_pb2

from ..._grpc_helpers.property_helper import mark_grpc_properties
from .base import (
    _ISOTROPIC_PROPERTY_UNAVAILABLE_MSG,
    _ORTHOTROPIC_PROPERTY_UNAVAILABLE_MSG,
    _ConstantPropertySet,
    _PolymorphicMixin,
    _PolymorphicPropertyKwargs,
    _VariablePropertySet,
)
from .property_helper import (
    constant_material_grpc_data_property,
    variable_material_grpc_data_property,
)

__all__ = [
    "ConstantStrainLimits",
    "VariableStrainLimits",
]


class _StrainLimitsMixin(_PolymorphicMixin):
    _DEFAULT_PB_PROPERTYSET_TYPE = material_pb2.OrthotropicStrainLimitsPropertySet
    _FIELD_NAME_SUFFIX_DEFAULT = "_orthotropic"
    _FIELD_NAME_SUFFIX_BY_PB_DATATYPE = {
        material_pb2.IsotropicStrainLimitsPropertySet: "_isotropic",
        material_pb2.OrthotropicStrainLimitsPropertySet: "_orthotropic",
    }


_ISOTROPIC_KWARGS: _PolymorphicPropertyKwargs = {
    "available_on_pb_type": material_pb2.IsotropicStrainLimitsPropertySet,
    "unavailable_msg": _ISOTROPIC_PROPERTY_UNAVAILABLE_MSG,
}
_ORTHOTROPIC_KWARGS: _PolymorphicPropertyKwargs = {
    "available_on_pb_type": material_pb2.OrthotropicStrainLimitsPropertySet,
    "unavailable_msg": _ORTHOTROPIC_PROPERTY_UNAVAILABLE_MSG,
}


@mark_grpc_properties
class ConstantStrainLimits(_StrainLimitsMixin, _ConstantPropertySet):
    """Constant strain limits material property set."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    @classmethod
    def from_isotropic_constants(
        cls,
        *,
        effective_strain: float = 0.0,
    ) -> Self:
        """Create an isotropic strain limits property set.

        Parameters
        ----------
        effective_strain:
            Effective strain limit.

        Returns
        -------
        :
            An isotropic strain limits property set.
        """
        obj = cls(
            _pb_object=cls._create_pb_object_from_propertyset_type(
                material_pb2.IsotropicStrainLimitsPropertySet
            )
        )
        obj.effective_strain = effective_strain
        return obj

    @classmethod
    def from_orthotropic_constants(
        cls,
        *,
        eXc: float = 0.0,
        eYc: float = 0.0,
        eZc: float = 0.0,
        eXt: float = 0.0,
        eYt: float = 0.0,
        eZt: float = 0.0,
        eSxy: float = 0.0,
        eSyz: float = 0.0,
        eSxz: float = 0.0,
    ) -> Self:
        """Create an orthotropic strain limits property set.

        Parameters
        ----------
        eXc:
            Compressive strain limit in material 1 direction.
        eYc:
            Compressive strain limit in material 2 direction.
        eZc:
            Compressive strain limit in out-of-plane direction.
        eXt:
            Tensile strain limit in material 1 direction.
        eYt:
            Tensile strain limit in material 2 direction.
        eZt:
            Tensile strain limit in out-of-plane direction.
        eSxy:
            Shear strain limit in-plane (:math:`e_{12}`).
        eSyz:
            Shear strain limit out-of-plane (:math:`e_{23}`).
        eSxz:
            Shear strain limit out-of-plane (:math:`e_{13}`).

        Returns
        -------
        :
            An orthotropic strain limits property set.
        """
        obj = cls(
            _pb_object=cls._create_pb_object_from_propertyset_type(
                material_pb2.OrthotropicStrainLimitsPropertySet
            )
        )
        obj.eXc = eXc
        obj.eYc = eYc
        obj.eZc = eZc
        obj.eXt = eXt
        obj.eYt = eYt
        obj.eZt = eZt
        obj.eSxy = eSxy
        obj.eSyz = eSyz
        obj.eSxz = eSxz
        return obj

    effective_strain = constant_material_grpc_data_property("effective_strain", **_ISOTROPIC_KWARGS)
    eXc = constant_material_grpc_data_property("eXc", **_ORTHOTROPIC_KWARGS)
    eYc = constant_material_grpc_data_property("eYc", **_ORTHOTROPIC_KWARGS)
    eZc = constant_material_grpc_data_property("eZc", **_ORTHOTROPIC_KWARGS)
    eXt = constant_material_grpc_data_property("eXt", **_ORTHOTROPIC_KWARGS)
    eYt = constant_material_grpc_data_property("eYt", **_ORTHOTROPIC_KWARGS)
    eZt = constant_material_grpc_data_property("eZt", **_ORTHOTROPIC_KWARGS)
    eSxy = constant_material_grpc_data_property("eSxy", **_ORTHOTROPIC_KWARGS)
    eSyz = constant_material_grpc_data_property("eSyz", **_ORTHOTROPIC_KWARGS)
    eSxz = constant_material_grpc_data_property("eSxz", **_ORTHOTROPIC_KWARGS)


@mark_grpc_properties
class VariableStrainLimits(_StrainLimitsMixin, _VariablePropertySet):
    """Variable strain limits material property set."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    effective_strain = variable_material_grpc_data_property("effective_strain", **_ISOTROPIC_KWARGS)
    eXc = variable_material_grpc_data_property("eXc", **_ORTHOTROPIC_KWARGS)
    eYc = variable_material_grpc_data_property("eYc", **_ORTHOTROPIC_KWARGS)
    eZc = variable_material_grpc_data_property("eZc", **_ORTHOTROPIC_KWARGS)
    eXt = variable_material_grpc_data_property("eXt", **_ORTHOTROPIC_KWARGS)
    eYt = variable_material_grpc_data_property("eYt", **_ORTHOTROPIC_KWARGS)
    eZt = variable_material_grpc_data_property("eZt", **_ORTHOTROPIC_KWARGS)
    eSxy = variable_material_grpc_data_property("eSxy", **_ORTHOTROPIC_KWARGS)
    eSyz = variable_material_grpc_data_property("eSyz", **_ORTHOTROPIC_KWARGS)
    eSxz = variable_material_grpc_data_property("eSxz", **_ORTHOTROPIC_KWARGS)
